#include <stdio.h>
#include "allegro.h"
#include "strings.h"

#include "trigtable.h"
#include "defs.h"
#include "game.h"
#include "level.h"
#include "spaceobj.h"
#include "explode.h"
#include "raptor.h"
#include "group.h"
#include "sound.h"
#include "pck.h"
#include "ebox.h"
#include "section.h"
#include <vector>
#include "hull_player.h"
#include "playerobj.h"
#include <aldumb.h>

#include <sys/time.h>

using namespace std;

GameRunnable::GameRunnable( bool dl, DATAFILE * _snd, char * fl, SpaceObject * player ):
level_speech( 0 ),
DRAW_LAND( dl ) {

	snd = _snd;

	finished = false;

	blend_trans = 0;

	((SAMPLE *)snd[ sustain ].dat)->priority = 1;
	((SAMPLE *)snd[ HULL_COLLIDE ].dat)->priority = 2;

	work = create_bitmap( screen_x, screen_y+Y_BLIT_OFFSET );
	//work = create_video_bitmap( screen_x, screen_y+Y_BLIT_OFFSET );
	if ( !work ) {
		printf("Supreme error with work\n");
		return;
	}
	clear( work );

	level = new LevelCreator( fl, player );

	expl = new ExplosionClass*[ MAX_EXPL ];
	for ( int q = 0; q < MAX_EXPL; q++ )
		expl[q] = NULL;
	
	land = NULL;
	if ( DRAW_LAND && !level->noExist() )
		LoadLand( land );

	last_explode = 0;

	fsec_keep = 25;
	fsec = 0;

	set_trans_blender( 0, 0, 0, 168 );

	view_port = screen_x / 4;

	player_shade = new int[ 40 ];
	player_shade_color = 0;
	blend_palette( player_shade, 20, makecol(85,0,0), makecol(255,0,0) );
	blend_palette( player_shade+20, 20, makecol(255,0,0), makecol(85,0,0) );

	dumb_file = NULL;
	dumb_player = NULL;

}


void GameRunnable::LoadLand(RLE_SPRITE *& who) {

	pckLoader *p = new pckLoader();
	char * ff = "1.pck";
	switch ( rnd( 3 ) ) {
		case 0  :       ff = "1.pck";break;
		case 1  :       ff = "2.pck";break;
		case 2  :       ff = "3.pck";break;
	}

	/*
	land2 = p->load( ff );
	if ( !land2 ){
		DRAW_LAND = false;
		printf("Supreme error with background\n");
	}
	*/

	BITMAP * ld = p->load( ff );
	if ( !ld ) {
		DRAW_LAND = FALSE;
		printf("Supreme error with land bitmap\n");
		delete p;
		return;
	}
	blit( ld, screen, 0, 0, 0, 0, GRAPHICS_X, GRAPHICS_Y );
	who = get_rle_sprite( ld );
	destroy_bitmap( ld );
	delete p;
}


void GameRunnable::draw_land( int y1 ) {

	draw_rle_sprite( work, land, view_port, y1-GRAPHICS_Y+Y_BLIT_OFFSET );
	draw_rle_sprite( work, land, view_port, y1+Y_BLIT_OFFSET );

	/*
	printf("Drawing land\n");
	draw_sprite( work, land2, view_port, Y_BLIT_OFFSET );
	printf("Stage 1\n");
	unsigned char * temp_line = land2->line[ land2->h - 1 ];
	printf("Stage 2\n");
	for ( int q = 1; q < land2->h; q++ )
		land2->line[q] = land2->line[q-1];
	printf("Stage 3\n");
	land2->line[ 0 ] = temp_line;
	printf("Made it\n");
	*/

}


void GameRunnable::HandleExplosion() {
	for ( int q = 0; q < MAX_EXPL; q++ )
		if( expl[q] != NULL )
			if ( expl[q]->update() ) {
				delete expl[q];
				expl[q] = NULL;
			}
}


void GameRunnable::HandleSpace( vector< SpaceObject * > * mv, vector< SpaceObject * > * fight, vector< SpaceObject * > * Ammo, Section * onscreen ) {

	for ( vector< SpaceObject * >::iterator it = mv->begin();
			it != mv->end(); ){
		if ( (*it)->MoveMe( Ammo, fight, onscreen, snd ) ) {
			SpaceObject * del = *it;
			onscreen->dispose( *it );
			it = mv->erase( it );
			delete del;
		} else ++it;
	}
}


void GameRunnable::CheckCollision( Section * good, Section * fight, SpaceObject * take, int sound, int vol ) {

	good->reset();
	fight->reset();
	
	vector< SpaceObject * > * good_space = good->getNext();
	vector< SpaceObject * > * fight_space = fight->getNext();

	while ( good_space != NULL && fight_space != NULL ){

		for ( vector< SpaceObject * >::iterator good_it = good_space->begin();
				good_it != good_space->end(); good_it++ ){

			for ( vector< SpaceObject * >::iterator fight_it = fight_space->begin();
					fight_it != fight_space->end(); fight_it++ ){
			
				SpaceObject * g_use = *good_it;
				SpaceObject * f_use = *fight_it;

				if ( f_use->getLife() > 0 && !g_use->haveCollide( f_use ) )
					if ( g_use->Collide( f_use ) ) {
						play_sound( snd, sound, vol );

						if ( g_use->Damage( f_use->Hurt() ) )
							g_use->Died(take,expl,MAX_EXPL,snd);
						if ( f_use->Damage( g_use->Hurt() ) )
							f_use->Died(take,expl,MAX_EXPL,snd);

						g_use->Collided( f_use, expl, MAX_EXPL );
						f_use->Collided( g_use, expl, MAX_EXPL );

						//g_use->addCollide( f_use );
						//f_use->addCollide( g_use );
					} //if

			} //for fight
			
		} //for good

		good_space = good->getNext();
		fight_space = fight->getNext();
		
	} //while
	
}


void GameRunnable::drawAll( int Land_count, vector< SpaceObject * > * draws , SpaceObject * player ) {
	
	//struct timeval * start, * end;
	//start = new timeval;
	//end = new timeval;
	
	if ( view_port < 0 ) view_port = 0;
	if ( view_port > screen_x - GRAPHICS_X ) view_port = screen_x - GRAPHICS_X;

	acquire_bitmap( work );

	//printf("Stage 1\n");
	//gettimeofday( start, NULL );
	if ( DRAW_LAND ) draw_land(Land_count%480);
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	//printf("Stage 2\n");
	//gettimeofday( start, NULL );
	if ( player != NULL ){
		if ( player->getLife() < 30 ) {
			rectfill( work, view_port, 0+Y_BLIT_OFFSET, view_port+GRAPHICS_X-1, 10+Y_BLIT_OFFSET, player_shade[player_shade_color] );
			rectfill( work, view_port, GRAPHICS_Y-10+Y_BLIT_OFFSET, view_port+GRAPHICS_X-1, GRAPHICS_Y-1+Y_BLIT_OFFSET, player_shade[player_shade_color] );
			if ( ++player_shade_color > 39 ) player_shade_color = 0;
		}
	
		int score = player->score;
		char * numnum = int2normal(score);
		char * final = append("Score ",numnum);
		raptor_font->rtext(work,view_port+11,GRAPHICS_Y-raptor_font->height()+Y_BLIT_OFFSET,makecol(255,128,0),final);
		free( numnum );
		free( final );


	}
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	vector< SpaceObject * > draw_real;

	/*
	for ( int q = 0; q <= MAX_PLANE; q++ )
		for ( vector< SpaceObject *>::iterator it = draws->begin();
				it != draws->end(); it++ )
			if ( (*it)->lookPlane() == q )
				(*it)->Shadow( work, data,q,Y_BLIT_OFFSET);
	*/
	
	//printf("Stage 3\n");
	//gettimeofday( start, NULL );
	for ( int q = 0; q <= MAX_PLANE; q++ )
		for ( vector< SpaceObject *>::iterator it = draws->begin();
				it != draws->end(); ){
			if ( (*it)->lookPlane() == q ){
				(*it)->Shadow( work, q,Y_BLIT_OFFSET);
				draw_real.push_back( *it );
				it = draws->erase( it );
			} else ++it;
		}
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	//printf("Stage 4\n");
	//gettimeofday( start, NULL );
	for ( vector< SpaceObject *>::iterator it = draw_real.begin(); it != draw_real.end(); it++ )
		(*it)->Draw( work, expl,MAX_EXPL,(*it)->lookPlane(),Y_BLIT_OFFSET);
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	/*
	for ( int q = 0; q <= MAX_PLANE; q++ )
		for ( vector< SpaceObject *>::iterator it = draw_real.begin();
				it != draw_real.end(); ){
			if ( (*it)->lookPlane() == q ){
				(*it)->Draw( work, data, expl,MAX_EXPL,q,Y_BLIT_OFFSET);
				it = draw_real.erase( it );
			} else ++it;
		}
	*/
	
	//printf("Stage 5\n");
	//gettimeofday( start, NULL );
	for ( int q = 0; q < MAX_EXPL; q++ ){
		if ( expl[q] != NULL )
			expl[q]->Draw( work, Y_BLIT_OFFSET );
		//if ( speed_counter != 0 ) q = MAX_EXPL;
	}
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	textprintf( work, font, 10 + view_port, 1+Y_BLIT_OFFSET, makecol(255,255,255), "FPS:%d", fsec_keep );
	textprintf( work, font, 10 + view_port, 10+Y_BLIT_OFFSET, makecol(255,255,255), "VERSION: 1.0.0" );
	if ( time(NULL) != timex ) {
		fsec_keep = fsec;
		fsec = 0;
		timex = time(NULL);
	}

	if ( player != NULL ) level_use = ( (PlayerObject *)player )->level;
	if ( level_speech < 120 ) {

		int * speech = new int[ 30 ];
		blend_palette( speech, 10, makecol( 255, 0, 120 ), makecol(150,220,20) );
		blend_palette( speech+10, 10, makecol(150, 220, 20), makecol(20,160,255) );
		blend_palette( speech+10+10, 10, makecol(20,160,255), makecol(255,0,120) );

		level_speech++;
		char * lose = int2str( level_use );
		char * dum = append("level ",lose);
		int vx = 300;
		int vy = 200;
		for ( int q = 0; q < length( dum ); q++ ) {
			int sy = (int)(vy + tsine[((level_speech+q*2)*12)%360]*8 );
			int sx = vx + q * 10;
			int col = 32 + ((level_speech+q*2)%20);
			col = speech[ (level_speech+q) % 30 ];
			textprintf( work, font, sx+view_port, sy+Y_BLIT_OFFSET, col, "%c", dum[q] );
		}
		free( dum );
		free( lose );
		delete[] speech;
	}
		

	if ( player != NULL ){
		PlayerHull * ph = (PlayerHull *)(player->hull);
		ph->showAttribute( work, ph->life, ph->max(0), view_port+GRAPHICS_X-10, Y_BLIT_OFFSET );
		ph->showAttribute( work, ph->shield, ph->max(1), view_port+1, Y_BLIT_OFFSET );
	}

	if ( finished ) {

		//draw_trans_sprite( work, blend_bright, 0, 0 );

	}
	
	//printf("Stage 6\n");
	//gettimeofday( start, NULL );
	acquire_screen();
	blit( work, screen, view_port, Y_BLIT_OFFSET, 0, 0, GRAPHICS_X, GRAPHICS_Y );
	//blit( work, screen, 0, Y_BLIT_OFFSET, 0, 0, screen_x, screen_y );
	//show_video_bitmap( work );
	//request_video_bitmap( work );
	release_screen();
	
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	//printf("Stage 7\n");
	//gettimeofday( start, NULL );
	if ( !DRAW_LAND ) {
		rectfill( work, view_port, Y_BLIT_OFFSET, view_port+GRAPHICS_X, GRAPHICS_Y+Y_BLIT_OFFSET, makecol(0,0,0) );
	}
	release_bitmap( work );
	//gettimeofday( end, NULL );
	//printf("Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

	//delete start;
	//delete end;

}


bool GameRunnable::end_level( vector< SpaceObject * > * PlayerTemp, double Land_real, double Land_speed ) {
	return true;
}


void GameRunnable::DeleteSpace( vector< SpaceObject *> * objs, Section * sec, SpaceObject * _player ) {


	for ( vector< SpaceObject * >::iterator it = objs->begin();
			it != objs->end(); ){

		if ( (*it)->getLife() <= 0 ){

			sec->dispose( *it );
			if ( (*it)->powerUp() ){
				if ( (*it)->group() != NULL ){
					if ( (*it)->group()->size() == 1 )
						if ( rnd(3)  ){
							int x = (*it)->actualx;
							int y = (*it)->actualy;
							objs->push_back( level->special(x,y,_player) );
						}
				}
			}
				
			SpaceObject * del = *it;
			it = objs->erase( it );
			delete del;

			
		} else ++it;
		
	}

}

void GameRunnable::DeleteAll( vector< SpaceObject * > * objs ){

	for ( vector<SpaceObject *>::iterator it = objs->begin();
			it != objs->end(); ){
		SpaceObject * del = *it;
		it = objs->erase( it );
		delete del;
	}
	
}


void GameRunnable::vectorAdd( vector< SpaceObject * > * stable, vector< SpaceObject * > * state ){

	for ( vector< SpaceObject * >::iterator it = state->begin(); 
			it != state->end(); it++ )
		stable->push_back( *it );
	
}

void GameRunnable::clearCollide( vector< SpaceObject * > * contain ) {

	for ( vector< SpaceObject * >::iterator it = contain->begin();
			it != contain->end(); it++ ){
		(*it)->clearCollide();
	}

}

DUH * GameRunnable::load_music(){


	char * music_file;
	DUH * who = NULL;
	switch( rnd( 4 ) ){
		case 0	:	music_file = data_file( "song1.xm" );break;
		case 1	:	music_file = data_file( "song2.xm" );break;
		case 2	:	music_file = data_file( "song3.xm" );break;
		case 3	:	music_file = data_file( "song4.xm" );break;
	}
	who = dumb_load_xm( music_file );
	free( music_file );

	return who;
	
}

int GameRunnable::RunGame( SpaceObject * player, int lev ) {

	if ( level->noExist() )
		return 2;

	fsec = 25;

	int Land_count = 0;
	double Land_speed = 2.7;
	double Land_real = Land_count;

	bool quit = false;
	int return_code = 0;

	speed_counter = 0;
	timex = time(NULL);

	vector< SpaceObject * > * GoodAmmo = new vector< SpaceObject * >;
	vector< SpaceObject * > * BadAmmo = new vector< SpaceObject * >;
	vector< SpaceObject * > * PlayerTemp = new vector< SpaceObject * >;
	vector< SpaceObject * > * Monster = new vector< SpaceObject * >;

	Section * player_section = new Section();
	Section * monster_section = new Section();
	Section * goodammo_section = new Section();
	Section * badammo_section = new Section();

	int score = player->score;
	PlayerTemp->push_back( player );

	HullObject * reserve = player->hull->copy();

	struct timeval * start, * end;
	start = new timeval;
	end = new timeval;

	//DUH * dumb_file = NULL;
	//AL_DUH_PLAYER * dumb_player = NULL;
	if ( music_vol != 0.0 ){
		dumb_file = load_music();
		dumb_player = al_start_duh( dumb_file, 2, 0, music_vol, 4096, 22050 );
	}
	
	#define SCROLL_VIEW 210

	while ( !quit ) {
		
		if ( key[KEY_L] ) {
			if ( DRAW_LAND )
				LoadLand( land );
			while ( key[KEY_L] );
		}

		if ( key[KEY_ESC] ){
			return_code = 4;
			quit = true;
		}

		//if ( key[KEY_R] ) quit = true;

		//puasy puasy
		if ( key[KEY_P] ) {
			BITMAP * pause = create_bitmap( screen_x, screen_y );
			clear( pause );
			set_trans_blender( 0, 0, 0, 128 );
			draw_trans_sprite( screen, pause, 0, 0 );
			raptor_font->rtext(screen,200,200,makecol(255,50,10),"Paused");
			while ( key[KEY_P] ){
				YIELD();
				if ( dumb_file != NULL )
					al_poll_duh( dumb_player );
			}
			/*
			while ( !key[KEY_P] ){
				YIELD();
				if ( dumb_file != NULL )
					al_poll_duh( dumb_player );
			}
			*/

			while ( !key[KEY_P] ){
				YIELD();
				if ( dumb_file != NULL )
					al_poll_duh( dumb_player );
			}
			while ( key[KEY_P] ){
				YIELD();
				if ( dumb_file != NULL )
					al_poll_duh( dumb_player );
			}
			set_trans_blender( 0, 0, 0, 168 );
			destroy_bitmap( pause );
			speed_counter = 0;
		}

		if ( dumb_file != NULL )
			al_poll_duh( dumb_player );

		bool did_something = false;
		while ( speed_counter > 0 ) {

			did_something = true;

			if ( finished && !quit )
				quit = end_level( PlayerTemp, Land_real, Land_speed );

			if ( !finished && level->done() ) {

				finished = Monster->empty();

			}

			level->create( Monster, lev, Land_real );
			//if ( player->getLife() < 30 )
			//	level->emergency( Monster );

			HandleExplosion();	  //this has to go here

			for ( int z = 0; z < USE_AMMO; z++ ) {

				if ( player->getLife() <= 0 ) {

					speed_counter = 0;
					for ( int play = 0; play < 60; ) {
		
						
						if ( dumb_file != NULL )
							al_poll_duh( dumb_player );

						bool get_dirty = false;
						while ( speed_counter > 0 ) {
							play++;
							get_dirty = true;
							//this has to go here
							HandleExplosion();
							for ( int ma = 0; ma < USE_AMMO; ma++ ) {

								goodammo_section->clear();
								badammo_section->clear();

								HandleSpace(GoodAmmo,Monster,GoodAmmo,goodammo_section);
								HandleSpace(BadAmmo,PlayerTemp,BadAmmo,badammo_section);

								CheckCollision(goodammo_section,monster_section,player,sustain,50);

								DeleteSpace(GoodAmmo,goodammo_section,NULL);
								DeleteSpace(BadAmmo,badammo_section,NULL);
								DeleteSpace(Monster,monster_section,player);

							}

							monster_section->clear();

							HandleSpace(Monster,PlayerTemp,BadAmmo,monster_section);

							DeleteSpace(Monster,monster_section,player);

							speed_counter--;
						}

						if ( get_dirty ) {
							fsec++;
							Land_real += Land_speed;
							if ( Land_real > 480 )
								Land_real -= 480;
							Land_count = (int)( Land_real + 1 ) % 480;

							if ( player->actualx < view_port + SCROLL_VIEW )
								view_port = player->actualx - SCROLL_VIEW;
							if ( player->actualx > view_port + GRAPHICS_X - SCROLL_VIEW )
								view_port += player->actualx - ( view_port + GRAPHICS_X - SCROLL_VIEW );

							vector< SpaceObject * > * draws = new vector< SpaceObject * >();

							vectorAdd( draws, GoodAmmo );
							vectorAdd( draws, BadAmmo );
							vectorAdd( draws, Monster );
							drawAll( Land_count, draws, NULL );

							delete draws;
						}

						if ( speed_counter == 0 ) YIELD();

					}	  //for

					//player->score -= 20000 * lev * 2 / 5;

					PlayerTemp->clear();
					speed_counter = 0;
					z = USE_AMMO;
					return_code = 1;
					quit = true;
				}		  //if

				goodammo_section->clear();
				badammo_section->clear();

				HandleSpace(GoodAmmo,Monster,GoodAmmo,goodammo_section);
				HandleSpace(BadAmmo,PlayerTemp,BadAmmo,badammo_section);
				
				clearCollide( GoodAmmo );
				clearCollide( BadAmmo );
				clearCollide( Monster );
				clearCollide( PlayerTemp );

				CheckCollision(badammo_section,player_section,player,sustain,50);
				CheckCollision(goodammo_section,monster_section,player,sustain,50);

				DeleteSpace(GoodAmmo,goodammo_section,NULL);
				DeleteSpace(BadAmmo,badammo_section,NULL);
				DeleteSpace(Monster,monster_section,player);

			}			  //for USE_AMMO

			player_section->clear();
			monster_section->clear();

			HandleSpace(PlayerTemp,Monster,GoodAmmo,player_section);
			HandleSpace(Monster,PlayerTemp,BadAmmo,monster_section);

			clearCollide( PlayerTemp );
			clearCollide( Monster );

			CheckCollision(monster_section,player_section,player,HULL_COLLIDE,50);
			DeleteSpace(Monster,monster_section,player);

			speed_counter--;
		}

		if ( did_something && !quit ) {
			fsec++;
			Land_real += Land_speed;
			if ( Land_real > 480 )
				Land_real -= 480;
			Land_count = (int)( Land_real ) % 480;
		
		
			if ( player->actualx < view_port + SCROLL_VIEW )
				view_port = player->actualx - SCROLL_VIEW;
			if ( player->actualx > view_port + GRAPHICS_X - SCROLL_VIEW )
				view_port += player->actualx - ( view_port + GRAPHICS_X - SCROLL_VIEW );
			/*
			if ( player->actualx > last_view ){

				int md = (player->actualx - last_view) / 2;
				
				//last_view++;
				//view_port++;
				last_view += md;
				view_port += md;
			}
			if ( player->actualx < last_view ){
				last_view--;
				view_port--;
			}
			*/
			/*
			int md = (player->actualx - last_view ) / 4;
			last_view += md;
			view_port += md;
			*/

			vector< SpaceObject * >* draws = new vector< SpaceObject * >();
			vectorAdd( draws, GoodAmmo );
			vectorAdd( draws, BadAmmo );
			vectorAdd( draws, Monster );
			vectorAdd( draws, PlayerTemp );

			//gettimeofday( start, NULL );
			drawAll( Land_count, draws, player );
			//gettimeofday( end, NULL );
			//printf("Total Time to draw = %0.1f\n", (double)(end->tv_usec - start->tv_usec) );

			delete draws;
		}

		if ( speed_counter == 0 )YIELD();

		//while ( speed_counter == 0 )yield_timeslice();

	}

	

	delete start;
	delete end;
	
	DeleteAll( Monster );
	DeleteAll( GoodAmmo );
	DeleteAll( BadAmmo );

	delete GoodAmmo;
	delete BadAmmo;
	delete Monster;
	delete PlayerTemp;

	delete goodammo_section;
	delete badammo_section;
	delete player_section;
	delete monster_section;

	if ( return_code != 0 ){
		player->giveHull( reserve );
		player->score = score;
	} else delete reserve;

	return return_code;

}


GameRunnable::~GameRunnable() {

	destroy_rle_sprite( land );
	if ( !level->noExist() ){
		BITMAP * fade_work = create_bitmap( GRAPHICS_X, GRAPHICS_Y );
		clear( fade_work );

		float mv = music_vol;
		blit( screen, work, 0, 0, 0, 0, GRAPHICS_X, GRAPHICS_Y );
		for ( int fade = 0; fade < 20; ) {
			if ( mv > 0 ) mv -= 0.1;
			al_duh_set_volume( dumb_player, mv );
			fade += speed_counter;
			if ( dumb_file != NULL )
				al_poll_duh( dumb_player );
			while ( speed_counter == 0 ){
				YIELD();
				if ( dumb_file != NULL )
					al_poll_duh( dumb_player );
			}
			if ( fade > 127 ) fade = 127;
			set_trans_blender( 0, 0, 0, fade );
			draw_trans_sprite( work, fade_work, 0, 0 );
			blit( work, screen, 0, 0, 0, 0, 640, 480 );
		}
		set_trans_blender(0,0,0,255);
		draw_trans_sprite(screen,fade_work,0,0);
		destroy_bitmap( fade_work );
	}

	if ( dumb_file != NULL ){
		al_stop_duh( dumb_player );
		unload_duh( dumb_file );
	}

	destroy_bitmap( work );

	if ( level )
		delete level;
	for ( int q = 0; q < MAX_EXPL; q++ )
		if ( expl[q] != NULL )
			delete expl[q];
	delete[] expl;
	delete[] player_shade;

}
